from elody.object_configurations.job_configuration import JobConfiguration


_config = JobConfiguration()
_create = _config.crud()["creator"]
_post_crud_hook = _config.crud()["post_crud_hook"]


def start_job(
    name,
    job_type,
    *,
    get_rabbit,
    get_user_context=None,
    user_email=None,
    parent_id=None,
    id_of_document_job_was_initiated_for=None,
    type_of_document_job_was_initiated_for=None,
) -> str:
    relations = []
    if parent_id:
        relations.append({"key": parent_id, "type": "hasParentJob"})
    if id_of_document_job_was_initiated_for and type_of_document_job_was_initiated_for:
        relations.append(
            {"key": id_of_document_job_was_initiated_for, "type": "isJobOf"}
        )

    job = _create(
        {
            "metadata": [
                {"key": "name", "value": name},
                {"key": "status", "value": "running"},
                {"key": "type", "value": job_type},
            ],
            "relations": relations,
            "type": "job",
        },
        get_user_context=get_user_context
        or (lambda: type("UserContext", (object,), {"email": user_email})()),
    )
    del job["computed_values"]["created_at"]

    _post_crud_hook(
        crud="create", document=job, parent_id=parent_id, get_rabbit=get_rabbit
    )
    __patch_document_job_was_initiated_for(
        id_of_document_job_was_initiated_for,
        type_of_document_job_was_initiated_for,
        get_rabbit,
    )
    return job["_id"]


def finish_job(
    id,
    id_of_document_job_was_initiated_for=None,
    type_of_document_job_was_initiated_for=None,
    *,
    get_rabbit,
):
    document = {
        "id": id,
        "patch": {
            "metadata": [{"key": "status", "value": "finished"}],
            "relations": ([] if id_of_document_job_was_initiated_for else []),
        },
    }
    _post_crud_hook(crud="update", document=document, get_rabbit=get_rabbit)
    __patch_document_job_was_initiated_for(
        id_of_document_job_was_initiated_for,
        type_of_document_job_was_initiated_for,
        get_rabbit,
    )


def fail_job(id, exception_message, *, get_rabbit):
    document = {
        "id": id,
        "patch": {
            "metadata": [
                {"key": "info", "value": exception_message},
                {"key": "status", "value": "failed"},
            ]
        },
    }
    _post_crud_hook(crud="update", document=document, get_rabbit=get_rabbit)


def __patch_document_job_was_initiated_for(id, type, get_rabbit):
    if id and type:
        document = {
            "document_info_job_was_initiated_for": {"id": id, "type": type},
            "patch": {"relations": [{"key": id, "type": "hasJob"}]},
        }
        _post_crud_hook(crud="update", document=document, get_rabbit=get_rabbit)



import re as regex

from copy import deepcopy
from elody.policies.permission_handler import (
    get_permissions,
    mask_protected_content_post_request_hook,
)
from flask import Request  # pyright: ignore
from inuits_policy_based_auth import BaseAuthorizationPolicy  # pyright: ignore
from inuits_policy_based_auth.contexts.policy_context import (  # pyright: ignore
    PolicyContext,
)
from inuits_policy_based_auth.contexts.user_context import (  # pyright: ignore
    UserContext,
)


class FilterGenericObjectsPolicyV2(BaseAuthorizationPolicy):
    def authorize(
        self, policy_context: PolicyContext, user_context: UserContext, request_context
    ):
        request: Request = request_context.http_request
        if not regex.match("^(/[^/]+/v[0-9]+)?/[^/]+/filter$", request.path):
            return policy_context

        if not isinstance(user_context.access_restrictions.filters, list):
            user_context.access_restrictions.filters = []
        type_filter, filters = self.__split_type_filter(
            user_context, deepcopy(request.json or [])
        )
        if not type_filter:
            policy_context.access_verdict = True
            return policy_context

        policy_context.access_verdict = False
        for role in user_context.x_tenant.roles:
            permissions = get_permissions(role, user_context)
            if not permissions:
                continue

            rules = [PostRequestRules]
            access_verdict = None
            for rule in rules:
                access_verdict = rule().apply(
                    type_filter["value"], filters, user_context, request, permissions
                )
                if access_verdict != None:
                    policy_context.access_verdict = access_verdict
                    if not policy_context.access_verdict:
                        break

            if policy_context.access_verdict:
                return policy_context

        return policy_context

    def __split_type_filter(self, user_context: UserContext, request_body: list):
        type_filter = None
        for filter in request_body:
            if filter["type"] == "type":
                type_filter = filter
            elif filter["type"] == "selection" and filter["key"] == "type":
                type_filter = filter
            elif item_types := filter.get("item_types"):
                type_filter = {
                    "type": "selection",
                    "key": "type",
                    "value": item_types,
                    "match_exact": True,
                }

        if not type_filter:
            if tenant_relation_type := user_context.bag.get("tenant_relation_type"):
                user_context.access_restrictions.filters.append(  # pyright: ignore
                    {
                        "type": "selection",
                        "key": tenant_relation_type,
                        "value": [
                            user_context.bag.get(
                                "tenant_defining_entity_id", user_context.x_tenant.id
                            )
                        ],
                        "match_exact": True,
                    }
                )
            return None, request_body

        try:
            request_body.remove(type_filter)
        except ValueError:
            pass
        return type_filter, request_body


class PostRequestRules:
    def apply(
        self,
        type_filter_values: str | list[str],
        filters: list[dict],
        user_context: UserContext,
        request: Request,
        permissions,
    ) -> bool | None:
        if request.method != "POST":
            return None

        if isinstance(type_filter_values, str):
            type_filter_values = [type_filter_values]

        type_filter_values_copy = deepcopy(type_filter_values)
        for type_filter_value in type_filter_values_copy:
            if type_filter_value not in permissions["read"].keys():
                type_filter_values.remove(type_filter_value)
                continue

            restrictions_grouped_by_index = {}
            schemas = permissions["read"][type_filter_value]
            for schema in schemas.keys():
                restrictions = schemas[schema].get("object_restrictions", {})
                for restricted_key, restricting_value in restrictions.items():
                    index, restricted_key = restricted_key.split(":")
                    key = f"{schema}|{restricted_key}"
                    if group := restrictions_grouped_by_index.get(index):
                        group["key"].append(key)
                    else:
                        restrictions_grouped_by_index.update(
                            {
                                index: {
                                    "key": [key],
                                    "value": restricting_value,
                                }
                            }
                        )

            # support false soft call read responses
            for filter in filters:
                key = filter.get("key", "")
                if isinstance(key, list):
                    key = ",".join(key)
                for restriction in restrictions_grouped_by_index.values():
                    if key not in ",".join(restriction["key"]):
                        continue
                    values = (
                        filter["value"]
                        if isinstance(filter["value"], list)
                        else [filter["value"]]
                    )
                    if not any(value in restriction["value"] for value in values):
                        return False

            for restriction in restrictions_grouped_by_index.values():
                user_context.access_restrictions.filters.append(  # pyright: ignore
                    {
                        "type": "selection",
                        "key": restriction["key"],
                        "value": restriction["value"],
                        "match_exact": True,
                    }
                )

        if len(type_filter_values) == 0:
            return False
        user_context.access_restrictions.post_request_hook = (
            mask_protected_content_post_request_hook(user_context, permissions)
        )
        return True



mediafile1 = {
    "_id": "2476c1a2-c323-499e-9e28-6d1562296f3f",
    "filename": "41695fc83cdf1af3166dc77cc0d41e7a-bever.jpg",
    "date_created": {"$date": "2024-11-28T08:33:47.479Z"},
    "version": 3,
    "thumbnail_file_location": "/iiif/3/41695fc83cdf1af3166dc77cc0d41e7a-bever.jpg/full/,150/0/default.jpg",
    "original_file_location": "/download/41695fc83cdf1af3166dc77cc0d41e7a-bever.jpg",
    "metadata": [
        {"key": "copyright_color", "value": "green", "lang": "en"},
        {"key": "copyright_color_calculation", "value": "automatic"},
    ],
    "identifiers": [
        "2476c1a2-c323-499e-9e28-6d1562296f3f",
        "41695fc83cdf1af3166dc77cc0d41e7a",
    ],
    "type": "mediafile",
    "sort": {
        "copyright_color": [{"value": "green", "lang": "en"}],
        "copyright_color_calculation": [{"value": "green"}],
    },
    "relations": [
        {
            "key": "c4d1f89a-2858-4dd6-ba6b-cb333b4a5d7f",
            "label": "hasMediafile",
            "type": "belongsTo",
            "is_primary": False,
            "is_primary_thumbnail": False,
            "metadata": [{"key": "order", "value": 2}],
            "sort": {"order": [{"value": 2}]},
        },
        {"key": "0645e24b-5223-4d91-bae1-2e1226033bef", "type": "hasLicense"},
        {"key": "4efa288e-ffa8-4850-bfc9-f942b63867fa", "type": "hasPhotographer"},
        {
            "key": "d067d0c8-f63c-435e-825d-1e4d2ede63f7",
            "label": "hasChild",
            "type": "hasChild",
            "metadata": [{"key": "order", "value": 4}],
            "sort": {"order": [{"value": 4}]},
        },
    ],
    "file_creation_date": {"$date": "2024-11-28T08:33:47.612Z"},
    "original_filename": "bever.jpg",
    "mimetype": "image/jpeg",
    "date_updated": {"$date": "2024-11-28T08:33:47.612Z"},
    "last_editor": "developers@inuits.eu",
    "img_height": 554,
    "img_width": 711,
}


mediafile2 = {
    "_id": "abfa72af-43ac-4c40-b969-0561eae4c122",
    "filename": "0494340e99df984a6a3d8dfd15df74a7-transcode-8w3809mt4j-1.jpg",
    "md5sum": "0494340e99df984a6a3d8dfd15df74a7",
    "transcode_file_location": "/download/0494340e99df984a6a3d8dfd15df74a7-transcode-8w3809mt4j-1.jpg",
    "thumbnail_file_location": "/iiif/3/0494340e99df984a6a3d8dfd15df74a7-transcode-8w3809mt4j-1.jpg/full/,150/0/default.jpg",
    "mimetype": "image/jpeg",
    "metadata": [{"key": "copyright_color", "value": "orange"}],
    "identifiers": [
        "abfa72af-43ac-4c40-b969-0561eae4c122",
        "0494340e99df984a6a3d8dfd15df74a7",
    ],
    "type": "mediafile",
    "sort": {"copyright_color": [{"value": "orange"}]},
    "relations": [
        {
            "key": "c4698e82-c2c3-4dda-b7d7-6d8556d34c6e",
            "label": "hasChild",
            "type": "belongsToParent",
            "metadata": [{"key": "order", "value": 4}],
            "sort": {"order": [{"value": 4}]},
        }
    ],
}



import pytest
import json

from elody.util import (
    parse_string_to_bool,
    interpret_flat_key,
    __flatten_dict_generator,
    get_raw_id,
    get_item_metadata_value,
    get_mimetype_from_filename,
    mediafile_is_public,
    read_json_as_dict,
    parse_url_unfriendly_string,
    CustomJSONEncoder,
)
from data import mediafile1, mediafile2
from datetime import datetime, timezone
from unittest.mock import mock_open, patch, MagicMock


def test_default_method_with_datetime():
    encoder = CustomJSONEncoder()
    dt = datetime(2023, 10, 1, 12, 0, 0, tzinfo=timezone.utc)
    result = encoder.default(dt)
    assert result == "2023-10-01T12:00:00+00:00"


def test_default_method_with_naive_datetime():
    encoder = CustomJSONEncoder()
    dt = datetime(2023, 10, 1, 12, 0, 0)
    result = encoder.default(dt)
    assert result == "2023-10-01T10:00:00+00:00"


def test_encode_method_with_non_datetime():
    encoder = CustomJSONEncoder()
    obj = {"key": "value"}
    result = encoder.encode(obj)
    expected_result = json.dumps(obj)
    assert result == expected_result


@pytest.mark.parametrize(
    "input_value, expected_output",
    [
        # Truthy values
        ("true", True),
        ("yes", True),
        ("1", True),
        ("Y", True),
        ("   true   ", True),
        # Falsy values
        ("false", False),
        ("no", False),
        ("0", False),
        ("N", False),
        ("   no   ", False),
        # Non-string inputs
        (123, 123),
        (None, None),
        (True, True),
        (False, False),
        # Edge cases
        ("", ""),  # Empty string
        ("maybe", "maybe"),  # Unexpected string
        ("random", "random"),  # Another unexpected string
    ],
)
def test_parse_string_to_bool(input_value, expected_output):
    assert parse_string_to_bool(input_value) == expected_output


def test_interpret_flat_key():
    object_lists = {"user": ["name", "email"], "address": ["street", "city"]}

    # Test case 1
    flat_key = "user.name"
    expected_output = [{"key": "user", "object_list": "user", "object_key": "name"}]
    assert interpret_flat_key(flat_key, object_lists) == expected_output

    # Test case 2
    flat_key = "user.email"
    expected_output = [{"key": "user", "object_list": "user", "object_key": "email"}]
    assert interpret_flat_key(flat_key, object_lists) == expected_output

    # Test case 3
    flat_key = "address.street"
    expected_output = [
        {"key": "address", "object_list": "address", "object_key": "street"}
    ]
    assert interpret_flat_key(flat_key, object_lists) == expected_output

    # Test case 4
    flat_key = "address.city"
    expected_output = [
        {"key": "address", "object_list": "address", "object_key": "city"}
    ]
    assert interpret_flat_key(flat_key, object_lists) == expected_output


def test_flatten_dict_generator():
    object_lists = {"items": "id"}

    data = {
        "name": "John",
        "address": {"street": "123 Main St", "city": "Anytown"},
        "items": [{"id": "1", "value": "item1"}, {"id": "2", "value": "item2"}],
        "tags": ["tag1", "tag2"],
    }

    expected_output = {
        "name": "John",
        "address.street": "123 Main St",
        "address.city": "Anytown",
        "items.1.id": "1",
        "items.1.value": "item1",
        "items.2.id": "2",
        "items.2.value": "item2",
        "tags": ["tag1", "tag2"],
    }

    result = dict(__flatten_dict_generator(object_lists, data, ""))
    assert result == expected_output


def test_flatten_dict_generator_nested():
    object_lists = {"items": "id"}

    data = {
        "user": {
            "name": "John",
            "address": {"street": "123 Main St", "city": "Anytown"},
            "items": [{"id": "1", "value": "item1"}, {"id": "2", "value": "item2"}],
            "tags": ["tag1", "tag2"],
        }
    }

    expected_output = {
        "user.name": "John",
        "user.address.street": "123 Main St",
        "user.address.city": "Anytown",
        "user.items.1.id": "1",
        "user.items.1.value": "item1",
        "user.items.2.id": "2",
        "user.items.2.value": "item2",
        "user.tags": ["tag1", "tag2"],
    }

    result = dict(__flatten_dict_generator(object_lists, data, ""))
    assert result == expected_output


def test_flatten_dict_generator_no_items():
    object_lists = {"items": "id"}

    data = {
        "name": "John",
        "address": {"street": "123 Main St", "city": "Anytown"},
        "items": [{"value": "item1"}, {"value": "item2"}],
        "tags": ["tag1", "tag2"],
    }

    expected_output = {
        "name": "John",
        "address.street": "123 Main St",
        "address.city": "Anytown",
        "items": [{"value": "item1"}, {"value": "item2"}],
        "tags": ["tag1", "tag2"],
    }

    result = dict(__flatten_dict_generator(object_lists, data, ""))
    assert result == expected_output


@pytest.mark.parametrize(
    "mediafile_modification, expected_output",
    [
        (lambda mf: mf, "2476c1a2-c323-499e-9e28-6d1562296f3f"),
        (lambda mf: mf.update({"_key": "test-key"}) or mf, "test-key"),
    ],
)
def test_get_raw_id(mediafile_modification, expected_output):
    modified_mediafile = mediafile_modification(mediafile1.copy())
    assert get_raw_id(modified_mediafile) == expected_output

    # Test for KeyError
    modified_mediafile.pop("_key", None)
    modified_mediafile.pop("_id", None)
    with pytest.raises(KeyError):
        get_raw_id(modified_mediafile)


@pytest.mark.parametrize(
    "metadata_key, expected_output",
    [
        ("copyright_color", "green"),
        ("copyright_color_calculation", "automatic"),
        ("not_existing", ""),
    ],
)
def test_get_item_metadata_value(metadata_key, expected_output):
    assert get_item_metadata_value(mediafile1, metadata_key) == expected_output


@pytest.mark.parametrize(
    "filename, expected_output",
    [
        ("test.jpg", "image/jpeg"),
        ("test.png", "image/png"),
        ("test.tif", "image/tiff"),
        ("test.mp4", "video/mp4"),
        ("test", "application/octet-stream"),
    ],
)
def test_get_mimetype_from_filename(filename, expected_output):
    assert get_mimetype_from_filename(filename) == expected_output


@pytest.mark.parametrize(
    "mediafile, expected_output",
    [
        (mediafile1, True),
        (mediafile2, False),
    ],
)
def test_mediafile_is_public(mediafile, expected_output):
    assert mediafile_is_public(mediafile) == expected_output


@pytest.fixture
def logger():
    return MagicMock()


def test_read_json_as_dict_success(logger):
    filename = "test.json"
    json_content = '{"key": "value"}'
    expected_output = {"key": "value"}

    with patch("builtins.open", mock_open(read_data=json_content)):
        result = read_json_as_dict(filename, logger)
        assert result == expected_output
        logger.error.assert_not_called()


def test_read_json_as_dict_file_not_found(logger):
    filename = "test.json"

    with patch("builtins.open", side_effect=FileNotFoundError):
        result = read_json_as_dict(filename, logger)
        assert result == {}
        logger.error.assert_called_once_with(f"Could not read {filename} as a dict: ")


def test_read_json_as_dict_invalid_json(logger):
    filename = "test.json"
    invalid_json_content = '{"key": "value"'  # Missing closing brace

    with patch("builtins.open", mock_open(read_data=invalid_json_content)):
        result = read_json_as_dict(filename, logger)
        assert result == {}
        logger.error.assert_called_once_with(
            f"Could not read {filename} as a dict: Expecting ',' delimiter: line 1 column 16 (char 15)"
        )


@pytest.mark.parametrize(
    "input_str, replace_char, return_unfriendly_chars, expected_output",
    [
        # Test cases with default behavior
        ("Hello World!", None, False, "Hello%20World%21"),
        ("abc/def", None, False, "abc%2Fdef"),
        # Test cases with replace_char specified
        ("Hello World!", "-", False, "Hello-World-"),
        ("abc/def", "*", False, "abc*def"),
        # Test cases with return_unfriendly_chars enabled
        ("Hello World!", None, True, ("Hello%20World%21", [" ", "!"])),
        ("abc/def", None, True, ("abc%2Fdef", ["/"])),
        ("Hello World!", "-", True, ("Hello-World-", [" ", "!"])),
        # Edge cases
        ("", None, False, ""),
        ("NoSpecialChars", None, False, "NoSpecialChars"),
    ],
)
def test_parse_url_unfriendly_string(
    input_str, replace_char, return_unfriendly_chars, expected_output
):
    result = parse_url_unfriendly_string(
        input_str,
        replace_char=replace_char,
        return_unfriendly_chars=return_unfriendly_chars,
    )
    assert result == expected_output



import re as regex

from configuration import get_object_configuration_mapper  # pyright: ignore
from elody.policies.helpers import get_content
from elody.policies.permission_handler import (
    get_permissions,
    handle_single_item_request,
    mask_protected_content_post_request_hook,
)
from flask import g, Request  # pyright: ignore
from inuits_policy_based_auth import BaseAuthorizationPolicy  # pyright: ignore
from inuits_policy_based_auth.contexts.policy_context import (  # pyright: ignore
    PolicyContext,
)
from inuits_policy_based_auth.contexts.user_context import (  # pyright: ignore
    UserContext,
)


class GenericObjectRequestPolicyV2(BaseAuthorizationPolicy):
    def authorize(
        self, policy_context: PolicyContext, user_context: UserContext, request_context
    ):
        request: Request = request_context.http_request
        if not regex.match("^(/[^/]+/v[0-9]+)?/[^/]+$", request.path):
            return policy_context

        for role in user_context.x_tenant.roles:
            permissions = get_permissions(role, user_context)
            if not permissions:
                continue

            rules = [PostRequestRules, GetRequestRules]
            access_verdict = None
            for rule in rules:
                access_verdict = rule().apply(user_context, request, permissions)
                if access_verdict != None:
                    policy_context.access_verdict = access_verdict
                    if not policy_context.access_verdict:
                        break

            if policy_context.access_verdict:
                return policy_context

        return policy_context


class PostRequestRules:
    def apply(
        self, user_context: UserContext, request: Request, permissions
    ) -> bool | None:
        if request.method != "POST":
            return None

        content = g.get("content") or request.json
        content = get_content(content, request, content)
        schema_type = get_object_configuration_mapper().get(content["type"]).SCHEMA_TYPE
        item = {**content, "schema": {"type": schema_type}}
        return handle_single_item_request(
            user_context, item, permissions, "create", content
        )


class GetRequestRules:
    def apply(
        self, user_context: UserContext, request: Request, permissions
    ) -> bool | None:
        if request.method != "GET":
            return None

        type_query_parameter = request.args.get("type")
        allowed_item_types = list(permissions["read"].keys())
        filters = []

        if type_query_parameter:
            if type_query_parameter not in allowed_item_types:
                return None
            type_permissions = permissions["read"][type_query_parameter]
            schemas = list(type_permissions.keys())
            if len(schemas) > 0:
                number_of_object_restrictions = len(
                    type_permissions[schemas[0]].get("object_restrictions", {}).keys()
                )
                for i in range(number_of_object_restrictions):
                    keys, values = [], []
                    for schema in schemas:
                        object_restrictions = type_permissions[schema].get(
                            "object_restrictions", {}
                        )
                        key = [
                            key
                            for key in object_restrictions.keys()
                            if key.startswith(f"{i}:")
                        ][0]
                        keys.append(f"{schema}|{key.split(':')[1]}")
                        values = object_restrictions[key]
                    filters.append(
                        {
                            "type": "selection",
                            "key": keys,
                            "value": values,
                            "match_exact": True,
                        }
                    )
            filters.insert(0, {"type": "type", "value": type_query_parameter})
        else:
            filters.append(
                {
                    "type": "selection",
                    "key": "type",
                    "value": allowed_item_types,
                    "match_exact": True,
                }
            )
            if tenant_relation_type := user_context.bag.get("tenant_relation_type"):
                filters.append(
                    {
                        "type": "selection",
                        "key": tenant_relation_type,
                        "value": [
                            "tenant:super",
                            user_context.bag.get(
                                "tenant_defining_entity_id", user_context.x_tenant.id
                            ),
                        ],
                        "match_exact": True,
                    }
                )

        user_context.access_restrictions.filters = filters
        user_context.access_restrictions.post_request_hook = (
            mask_protected_content_post_request_hook(user_context, permissions)
        )
        return True
